# remove when dropping Python 3.7-3.9 support
from __future__ import annotations

import io
import tokenize
from collections.abc import Iterable
from typing import NamedTuple, TextIO

from mypy_upgrade.parsing import MypyError


class UnsilenceableRegion(NamedTuple):
    """A region within a source code that cannot be silenced by an inline
    comment.

    Attributes:
        start: a 2-tuple representing the start of the unsilenceable region
            whose first entry is the start line (1-indexed) and whose second
            entry is the start column offset.
        end: a 2-tuple representing the end of the unsilenceable region
            whose first entry is the end line (1-indexed) and whose second
            entry is the end column offset.

        When start[0] = end[0], it is interpreted that the Unsilenceable
        region is an explicitly continued line.
    """

    start: tuple[int, int]  # line, column
    end: tuple[int, int]  # line, column


def split_code_and_comment(line: str) -> tuple[str, str]:
    """Split a line of code into the code part and comment part."""
    reader = io.StringIO(line).readline

    try:
        comment_tokens = []
        for t in tokenize.generate_tokens(reader):
            if t.type == tokenize.COMMENT:
                comment_tokens.append(t)

        if not comment_tokens:
            return line.rstrip(), ""

        comment_token = comment_tokens[0]
        python_code = line[0 : comment_token.start[1]]
        python_comment = line[comment_token.start[1] :]
    except tokenize.TokenError:  # compromise for multi-line statements
        comment_start = line.find("#")
        if comment_start >= 0:
            python_code = line[0:comment_start]
            python_comment = line[comment_start:]
        else:
            python_code, python_comment = (line, "")

    return python_code.rstrip(), python_comment


def find_unsilenceable_regions(stream: TextIO) -> list[UnsilenceableRegion]:
    """Find the regions encapsulated by line continuation characters or
    by multiline strings

    Args:
        stream: A text stream.

    Returns:
        A list of UnsilenceableRegion objects.

        Multiline strings are represented by UnsilienceableRegion objects
        whose first entries in their `start` and `end` attributes are
        different. Explicitly continued lines are represented by
        UnsilienceableRegion objects whose first entries in their `start` and
        `end` attributes are the same.
    """
    all_lines = list(tokenize.generate_tokens(stream.readline))
    unsilenceable_regions = []
    for token in all_lines:
        if (
            token.start[0] != token.end[0]
            and token.exact_type == tokenize.STRING
        ):
            region = UnsilenceableRegion(token.start, token.end)
            unsilenceable_regions.append(region)

    comments = [t for t in all_lines if t.exact_type == tokenize.COMMENT]

    for token in all_lines:
        if token.line.rstrip("\r\n").endswith("\\") and not any(
            comment.line == token.line for comment in comments
        ):
            start = token.end[0], 0
            end = token.end[0], len(token.line)
            region = UnsilenceableRegion(start, end)
            unsilenceable_regions.append(region)

    return unsilenceable_regions


def find_safe_end_line(
    error: MypyError, unsilenceable_regions: Iterable[UnsilenceableRegion]
) -> int:
    """Find a syntax-safe line on which to place an error suppression comment
    for the given error.

    Args:
        error: a `MypyError` for which a type error suppression comment is to
            placed.
        unsilenceable_regions: an `Iterable` in which each entry is an
            `UnsilenceableRegion`.

    Returns:
        An integer representing a safe line on which to place an error
        suppression comment if it exists. If no safe line exists, this method
        returns -1.
    """
    new_line = None
    new_col_offset = None
    for region in unsilenceable_regions:
        # It is safe to comment the last line of a multiline string
        if error.line_no == region.end[0] and region.start[0] != region.end[0]:
            continue

        # It is indeterminant whether an error within an UnsilenceableRegion
        # can be suppressed if its column number is unknown
        if (
            error.col_offset is None
            and region.start[0] <= error.line_no <= region.end[0]
        ):
            return -1

        if (
            error.col_offset is not None
            and region.start <= (error.line_no, error.col_offset) <= region.end
        ):
            return -1

        # Error precedes same line multiline string
        if (
            error.line_no == region.start[0]
            and region.start[0] != region.end[0]
        ):
            new_line = region.end[0]
            new_col_offset = int(region.end[1])

    if new_line is None:
        return error.line_no

    new_error = MypyError(
        error.filename,
        new_line,
        new_col_offset,
        error.message,
        error.error_code,
    )
    return find_safe_end_line(new_error, unsilenceable_regions)


def correct_line_numbers(
    stream: TextIO, errors: Iterable[MypyError]
) -> tuple[list[MypyError], list[MypyError]]:
    """Correct the line numbers of MypyErrors considering multiline statements.

    Args:
        stream: A text stream from which the line numbers of provided errors
            are to be corrected.
        errors: The errors whose line numbers are to be corrected.

    Returns:
        A 2-tuple whose first entry is a list in which each entry is a
        `MypyError` from `errors` for which type suppression comments can be
        added (with line numbers corrected) and whose second entry is a list
        of each `MypyError` that cannot be silenced.
    """
    unsilenceable_regions = find_unsilenceable_regions(stream)
    line_corrected_errors = []
    unsilenceable_errors = []
    for error in errors:
        end_line = find_safe_end_line(error, unsilenceable_regions)

        if end_line == -1:
            unsilenceable_errors.append(error)
        else:
            line_corrected_errors.append(
                MypyError(
                    error.filename,
                    end_line,
                    error.col_offset,
                    error.message,
                    error.error_code,
                )
            )

    return line_corrected_errors, unsilenceable_errors



# remove when dropping Python 3.7-3.9 support
from __future__ import annotations

import re
from typing import NamedTuple, TextIO


class MypyError(NamedTuple):
    filename: str
    line_no: int
    col_offset: int | None
    message: str
    error_code: str

    @staticmethod
    def filename_and_line_number(error: MypyError) -> tuple[str, int]:
        return error.filename, error.line_no


def parse_mypy_report(
    report: TextIO,
) -> list[MypyError]:
    """Parse a mypy error report from stdin.

    Args:
        report: a text stream from which to read the mypy typing report
    Returns:
        A sorted list of MypyErrors. Elements are sorted by their filename
        and line_no attributes.

        Example::

            >> report = pathlib.Path('mypy_report.txt').open(encoding='utf-8')
            >> errors = parse_report(report)
            >> module, col_offset, line_no, message , error_code= errors[0]
    """
    info = re.compile(
        r"^(?P<filename>[^:]+):(?P<line_no>\d+)(:(?P<col_offset>\d+))?"
        r"(:\d+:\d+)?: error: (?P<message>.+)\s+(\[(?P<error_code>.+)\])?"
    )
    errors = []

    for line in report:
        error = info.match(line.strip())
        if error:
            error_code = error.group("error_code") or ""
            filename, message = error.group("filename", "message")
            line_no = int(error.group("line_no"))
            if error.group("col_offset"):
                col_offset = int(error.group("col_offset"))
            else:
                col_offset = None

            errors.append(
                MypyError(
                    filename,
                    line_no,
                    col_offset,
                    message.strip(),
                    error_code,
                )
            )

    return sorted(errors, key=MypyError.filename_and_line_number)


def description_to_type_ignore(description: str) -> tuple[str, ...]:
    type_ignore_re = re.compile(
        r"type\s*:\s*ignore\s*(\[(?P<error_code>[a-z, \-]+)\])?"
    )
    # Extract unused type ignore error codes from error description
    match = type_ignore_re.search(description)
    if match:
        error_codes = match.group("error_code")
        if error_codes:
            # Separate and trim
            return tuple(code.strip() for code in error_codes.split(","))

    return ()



# remove when dropping Python 3.7-3.9 support
from __future__ import annotations

import re
from collections.abc import Iterable


def add_type_ignore_comment(comment: str, error_codes: list[str]) -> str:
    """Add type ignore comment with error codes to in-line comment.

    Args:
        comment: A comment in which to add a type ignore comment.
        error_codes: The error codes to add to the type ignore comment.

    Returns:
        A copy of the comment with a "type: ignore[error-code]" comment
    """
    old_type_ignore_re = re.compile(
        r"type\s*:\s*ignore\[(?P<error_code>[a-z, \-]+)\]"
    )

    # Handle existing "type: ignore[error_code]" comments
    match = old_type_ignore_re.search(comment)
    if match:
        old_error_codes = set(
            match.group("error_code").replace(" ", "").split(",")
        )
        error_codes.extend(e for e in old_error_codes if e not in error_codes)
        comment = old_type_ignore_re.sub("", comment)

        # Check for other comments; otherwise, remove comment
        if not re.search(r"[^#\s]", comment):
            comment = ""
        else:
            comment = f' # {comment.lstrip("# ")}'
    elif comment:
        # format comment
        comment = f' # {comment.lstrip("# ")}'

    sorted_error_codes = ", ".join(sorted(error_codes))

    return f"# type: ignore[{sorted_error_codes}]{comment}"


def format_type_ignore_comment(comment: str) -> str:
    """Remove excess whitespace and commas from `"type: ignore"` comments."""
    type_ignore_re = re.compile(
        r"type\s*:\s*ignore(\[(?P<error_codes>[a-z, \-]+)\])?"
    )
    match = type_ignore_re.search(comment)

    # Format existing error codes
    if match:
        error_codes = match.group("error_codes")
        if error_codes:
            pruned_error_codes = []
            for code in error_codes.split(","):
                pruned_code = code.strip()
                if pruned_code:
                    pruned_error_codes.append(pruned_code)

            formatted_comment = comment.replace(
                error_codes, ", ".join(pruned_error_codes)
            )
            if pruned_error_codes:
                return formatted_comment

            # Format again if there are no error codes
            return format_type_ignore_comment(formatted_comment)

    # Delete "type: ignore", "type: ignore[]"
    formatted_comment = re.sub(r"type\s*:\s*ignore\s*(\[\])?", "", comment)

    # Return empty string if nothing is left in the comment
    if not re.search(r"[^#\s]", formatted_comment):
        return ""

    return formatted_comment


def remove_unused_type_ignore(
    comment: str, codes_to_remove: Iterable[str]
) -> str:
    """Remove specified error codes from a comment string.

    Args:
        comment: a string whose "type: ignore" codes are to be removed.
        codes_to_remove: an iterable of strings which represent mypy error
            codes. If this iterable is length zero, the entire "type: ignore"
            comment is removed.

    Returns:
        A copy of the original string with the specified error codes removed.
    """
    if codes_to_remove:
        pruned_comment = comment
        for code_to_remove in codes_to_remove:
            pruned_comment = pruned_comment.replace(code_to_remove, "")
    else:
        pruned_comment = re.sub(
            r"type\s*:\s*ignore(\[[a-z, \-]+\])?", "", comment
        )

    return pruned_comment



"""This defines a tool to silence mypy errors using in-line comments.
"""
# remove when dropping Python 3.7-3.9 support
from __future__ import annotations

import argparse
import itertools
import pathlib
import shutil
import sys
import textwrap
from typing import NamedTuple

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from mypy_upgrade.__about__ import __version__
from mypy_upgrade.filter import filter_mypy_errors
from mypy_upgrade.parsing import (
    MypyError,
    parse_mypy_report,
)
from mypy_upgrade.silence import silence_errors
from mypy_upgrade.utils import correct_line_numbers
from mypy_upgrade.warnings import (
    MISSING_ERROR_CODES,
    TRY_SHOW_ABSOLUTE_PATH,
    create_not_silenced_errors_warning,
)


class MypyUpgradeResult(NamedTuple):
    silenced: tuple[MypyError, ...]
    not_silenced: tuple[MypyError, ...]
    messages: tuple[str, ...]


def _create_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="mypy-upgrade",
        description="""
Place in-line comments into files to silence mypy errors.
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples
--------

# Pyre-like invocation
$ mypy --strict -p package | mypy-upgrade

# Use saved report file
$ mypy --strict -p package > mypy_report.txt
$ mypy-upgrade --report mypy_report.txt

# Only silence errors in package/module
$ mypy --strict -p package > mypy_report.txt
$ mypy-upgrade -p package.subpackage -m package.module --report mypy_report.txt

# Only silence errors in file/directory
$ mypy --strict -p package > mypy_report.txt
$ mypy-upgrade --report mypy_report.txt package/module.py doc
""",
    )
    parser.add_argument(
        "-m",
        "--module",
        default=[],
        dest="modules",
        metavar="MODULE",
        action="append",
        help="Silence errors from the provided (importable) module. "
        "This flag may be repeated multiple times.",
    )
    parser.add_argument(
        "-p",
        "--package",
        default=[],
        dest="packages",
        metavar="PACKAGE",
        action="append",
        help="Silence errors from the provided (importable) package. "
        "This flag may be repeated multiple times.",
    )
    parser.add_argument(
        "-r",
        "--report",
        type=pathlib.Path,
        help="""
        The path to a text file containing a mypy type checking report. If not
        specified, input is read from standard input. Defaults to stdin.
        """,
    )
    parser.add_argument(
        "-d",
        "--description-style",
        default="none",
        choices=["full", "none"],
        help="""
        Specify the style in which mypy error descriptions are expressed in the
        error suppression comment. Defaults to "none".
        """,
    )
    parser.add_argument(
        "--fix-me",
        default="FIX ME",
        help="""
        Specify a custom 'Fix Me' message to be placed after the error
        suppression comment. Pass " " to omit a 'Fix Me' message altogether.
        Defaults to "FIX ME".
        """,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help=(
            "Control the verbosity."
            "0: Only warnings are printed."
            "1: Print detailed warnings, a short summary of silenced errors, "
            "and a detailed list of errors that were not silenced."
            "2: Print detailed warnings, a detailed list of silenced errors, "
            "and a detailed list of errors that were not silenced. Defaults "
            "to 0."
        ),
    )
    parser.add_argument(
        "-V",
        "--version",
        default=False,
        action="store_const",
        const=True,
        help="Print the version.",
    )
    parser.add_argument(
        "--suppress-warnings",
        default=False,
        action="store_const",
        const=True,
        help="Suppress all warnings.",
    )
    parser.add_argument(
        "files",
        default=[],
        nargs="*",
        help="Silence errors from the provided files/directories.",
    )
    return parser


def mypy_upgrade(
    report: pathlib.Path | None,
    packages: list[str],
    modules: list[str],
    files: list[str],
    description_style: Literal["full", "none"],
    fix_me: str,
) -> MypyUpgradeResult:
    """Main logic for application.

    Args:
        report: an optional `pathlib.Path` pointing to the the mypy error
            report. If `None`, the report is read from the standard input.
        packages: a list of strings representing the packages in which to
            silence errors.
        modules: a list of strings representing the modules in which to
            silence errors.
        files: a list of strings representing the files in which to
            silence errors.
        description_style: a string specifying the style of error descriptions
            appended to the end of error suppression comments. A value of
            "full" appends the complete error message. A value of "none"
            does not append anything.
        fix_me: a string specifying the 'Fix Me' message in type error
            suppresion comments. Pass "" to omit a 'Fix Me' message
            altogether. All trailing whitespace will be trimmed.

    Returns:
        A `MypyUpgradeResult` object. The errors that are silenced via type
        checking suppression comments are stored in the `silenced` attribute.
        Those that are unable to be silenced are stored in the `not_silenced`
        attribute. If a `FileNotFoundError` is raised while reading a file in
        which an error is to be silenced or `mypy-upgrade`-related warnings
        are raised during execution are stored in the `messages` attribute.
    """
    if report is not None:
        with pathlib.Path(report).open(encoding="utf-8") as file:
            errors = parse_mypy_report(file)
    else:
        errors = parse_mypy_report(sys.stdin)

    filtered_errors = filter_mypy_errors(errors, packages, modules, files)

    messages = []
    not_silenced: list[MypyError] = []
    silenced: list[MypyError] = []
    for filename, filename_grouped_errors in itertools.groupby(
        filtered_errors, key=lambda error: error.filename
    ):
        try:
            with pathlib.Path(filename).open(encoding="utf-8") as f:
                safe_to_silence, unsafe_to_silence = correct_line_numbers(
                    f, filename_grouped_errors
                )
                f.seek(0)
                lines = f.readlines()
        except FileNotFoundError:
            messages.append(
                TRY_SHOW_ABSOLUTE_PATH.replace("{filename}", filename)
            )
            return MypyUpgradeResult(
                tuple(silenced), tuple(not_silenced), tuple(messages)
            )

        not_silenced.extend(unsafe_to_silence)

        for line_number, line_grouped_errors in itertools.groupby(
            safe_to_silence, key=lambda error: error.line_no
        ):
            lines[line_number - 1] = silence_errors(
                lines[line_number - 1],
                line_grouped_errors,
                description_style,
                fix_me,
            )

        with pathlib.Path(filename).open(mode="w", encoding="utf-8") as f:
            _ = f.write("".join(lines))

        if safe_to_silence:
            silenced.extend(safe_to_silence)

    if any(error.error_code is None for error in silenced + not_silenced):
        messages.append(MISSING_ERROR_CODES)

    return MypyUpgradeResult(
        tuple(silenced), tuple(not_silenced), tuple(messages)
    )


def print_results(
    results: MypyUpgradeResult, options: dict[str, int | bool]
) -> None:
    """Print the results contained in a `MypyUpgradeResult` object.

    Args:
        results: a `MypyUpgradeResult` object.
        verbosity: an integer specifying the verbosity.
    """
    width = min(79, shutil.get_terminal_size(fallback=(140, 0)).columns)

    def fill_(text: str) -> str:
        return textwrap.fill(text, width=width)

    if results.not_silenced and not options["suppress_warnings"]:
        not_silenced_warning = create_not_silenced_errors_warning(
            results.not_silenced, options["verbosity"]
        )
        print(" WARNING ".center(width, "-"))  # noqa: T201
        print(fill_(not_silenced_warning))  # noqa: T201
        print()  # noqa: T201

    if not options["suppress_warnings"]:
        for message in results.messages:
            print(" WARNING ".center(width, "-"))  # noqa: T201
            print(fill_(message))  # noqa: T201
            print()  # noqa: T201

    if options["verbosity"] == 1:
        num_files = len({err.filename for err in results.silenced})
        num_silenced = len(results.silenced)
        text = fill_(
            f"{num_silenced} error{'' if num_silenced == 1 else 's'} "
            f"silenced across {num_files} file{'' if num_files == 1 else 's'}."
        )
        print(text)  # noqa: T201
    elif options["verbosity"] > 1:
        if results.silenced:
            print(  # noqa: T201
                f" ERRORS SILENCED ({len(results.silenced)}) ".center(
                    width, "-"
                )
            )
            for error in results.silenced:
                print(  # noqa: T201
                    f"{error.error_code}: {error.filename} ({error.line_no})"
                )
        if results.not_silenced:
            print(  # noqa: T201
                f" ERRORS NOT SILENCED ({len(results.not_silenced)}) ".center(
                    width, "-"
                )
            )
            for error in results.not_silenced:
                print(  # noqa: T201
                    f"{error.error_code}: {error.filename} ({error.line_no})"
                )


def main() -> None:
    """An interface to `mypy_upgrade` from the command-line."""
    parser = _create_argument_parser()
    args = parser.parse_args()
    if args.version:
        print(f"mypy-upgrade {__version__}")  # noqa: T201
        return None

    try:
        results = mypy_upgrade(
            args.report,
            args.packages,
            args.modules,
            args.files,
            args.description_style,
            args.fix_me.rstrip(),
        )
    except FileNotFoundError as error:
        if error.filename == args.report:
            print(  # noqa: T201
                f"Aborting: Unable to find report {args.report}"
            )
        else:
            raise

    print_results(results, args.verbose)



# remove when dropping Python 3.7-3.9 support
from __future__ import annotations

import importlib.abc
import pathlib
from importlib import util

from mypy_upgrade.parsing import MypyError


def get_module_paths(modules: list[str]) -> list[pathlib.Path | None]:
    """Determine file system paths of given modules/packages.

    Args:
        modules: a list of strings representing (importable) modules.

    Returns:
        A list (of the same length as the input list) of pathlib.Path objects
        corresponding to the given modules. If a path is not found for a
        module, the corresponding entry in the output is ``None``.

    Raises:
        NotImplementedError: Uncountered an unsupported module type.
    """
    paths: list[pathlib.Path | None] = []
    for module in modules:
        spec = util.find_spec(module)
        if spec is None:
            paths.append(None)
        else:
            loader = spec.loader
            if isinstance(loader, importlib.abc.ExecutionLoader):
                module_path = pathlib.Path(loader.get_filename(module))
                if loader.is_package(module):
                    module_path = module_path.parent
            else:
                msg = "Uncountered an unsupported module type."
                raise NotImplementedError(msg)
            paths.append(module_path)

    return paths


def filter_mypy_errors(
    errors: list[MypyError],
    packages: list[str],
    modules: list[str],
    files: list[str],
) -> list[MypyError]:
    """Select errors based on specified packages, modules, files.

    Args:
        errors: a list of MypyErrors.
        packages: a list of strings specifying packages to be included.
        modules: a list of strings specifying modules to be included.
        files: a list of strings specifying files to be included.

    Returns:
        A list of MypyErrors including only those in either ``packages``,
        ``modules``, or ``files``.
    """
    if len(packages + modules + files) == 0:
        return errors

    package_paths = [p for p in get_module_paths(packages) if p is not None]
    module_paths = [m for m in get_module_paths(modules) if m is not None]
    file_paths = [pathlib.Path(f).resolve() for f in files]
    paths = package_paths + module_paths + file_paths
    selected = []
    for error in errors:
        module_path = pathlib.Path(error.filename).resolve()
        # ! Use Path.is_relative_to when dropping Python 3.7-3.8 support
        should_include = any(
            path in module_path.parents or path == module_path
            for path in paths
        )

        if should_include:
            selected.append(error)

    return selected



# remove when dropping Python 3.7-3.9 support
from __future__ import annotations

import sys
from collections.abc import Iterable

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from mypy_upgrade.editing import (
    add_type_ignore_comment,
    format_type_ignore_comment,
    remove_unused_type_ignore,
)
from mypy_upgrade.parsing import MypyError, description_to_type_ignore
from mypy_upgrade.utils import split_code_and_comment


def silence_errors(
    line: str,
    errors: Iterable[MypyError],
    description_style: Literal["full", "none"],
    fix_me: str,
) -> str:
    """Silences the given error on a line with an error code-specific comment.

    Args:
        line: a string containing the line.
        error: an `Iterable` in which each entry is a `MypyError` to be
            silenced.
        description_style: a string specifying the style of the description of
            errors.
        fix_me: a string specifying a "fix me" message to be appended after the
            silencing comment.
    Returns:
        The line with a type error suppression comment.
    """
    unused_ignore = None
    error_codes = []
    descriptions = []
    for error in errors:
        if error.error_code == "unused-ignore":
            unused_ignore = error  # there should only be one
        elif error.error_code not in error_codes:
            error_codes.append(error.error_code)
            descriptions.append(error.message)

    python_code, comment = split_code_and_comment(line.rstrip())

    if unused_ignore:
        codes_to_remove = description_to_type_ignore(unused_ignore.message)
        pruned_comment = remove_unused_type_ignore(comment, codes_to_remove)
        cleaned_comment = format_type_ignore_comment(pruned_comment)
    else:
        cleaned_comment = comment

    if error_codes:
        final_comment = add_type_ignore_comment(
            cleaned_comment,
            error_codes,
        )
    else:
        final_comment = cleaned_comment

    updated_line = f"{python_code}  {final_comment}".rstrip()

    if fix_me:
        updated_line += f" # {fix_me}"

    if description_style == "full" and descriptions:
        updated_line += f" # {', '.join(descriptions)}"

    return updated_line + "\n"



from __future__ import annotations

from mypy_upgrade.parsing import MypyError

MISSING_ERROR_CODES = (
    "Not all errors in the mypy type checking report have error "
    "codes. As a result, mypy-upgrade can only suppress these errors "
    "with non-code-specific # type: ignore comments (which will still "
    "raise errors when running mypy with --strict enabled). If you "
    "would like mypy-upgrade to silence errors with code-specific "
    "comments, please run mypy with --show-error-codes enabled. "
    "If you would like to suppress this warning, use the "
    "--allow-no-error-codes flag for mypy-upgrade."
)

TRY_SHOW_ABSOLUTE_PATH = (
    "Unable to find file {filename}. This may be due to running"
    "mypy in a different directory than mypy-upgrade. Please try "
    "running mypy with the --show-absolute-path flag set."
)


def create_not_silenced_errors_warning(
    not_silenced: tuple[MypyError, ...], verbosity: int = 0
) -> str:
    """Create a warning for the user about errors that were not silenced.

    Args:
        not_silenced: a list of MypyErrors that were not silenced.
        verbosity: an integer indicating the verbosity level for printing

    Returns:
        A string containing the warning.
    """
    if len(not_silenced) == 0:
        with_column_numbers = False
    else:
        with_column_numbers = all(
            err.col_offset is not None for err in not_silenced
        )

    num_not_silenced = len(not_silenced)
    verb = "error was not" if num_not_silenced == 1 else "errors were not"

    warning_stem = (
        f"{num_not_silenced} {verb} not silenced due to syntax "
        "limitations.\n\n"
    )
    if not with_column_numbers:
        warning_suffix = (
            "Consider using the --show-column-numbers flag when "
            "generating the mypy type checking report."
        )
    else:
        if verbosity < 1:
            verbose_suggestion = (
                "\n\nRun mypy-upgrade in verbose mode (option -v) to get a "
                "full print out of affected type checking errors."
            )
        else:
            verbose_suggestion = ""

        warning_suffix = (
            "Line continuation characters and multiline (f-strings) are "
            "common culprits. Possible resolutions include: 1) "
            "formatting the affected code with a PEP 8-compliant "
            "formatter (e.g., black), 2) refactoring the affected code "
            "to avoid line continutation characters or multiline "
            "f-strings, and 3) resolving the errors."
            f"{verbose_suggestion}"
        )

    return warning_stem + warning_suffix
